# put this as 'pasef-hooks.py' inside an otofControl method (.m) folder.
#
# must define a 'PasefHooks' class.
#
# file will be re-loaded and class re-instantiated whenever loading a method in otofControl.

import logging
import sqlite3, os
import traceback

import numpy as np
from kafka import KafkaConsumer, KafkaProducer
import sys
import time
from datetime import datetime
import ctypes
import threading
import sys
import json
import math
import subprocess

# Handles Plotting the duty cycle
import matplotlib

matplotlib.use('TkAgg')
import numpy as np, matplotlib.pyplot as plt

logging.info('Loading my PASEF hooks ...')

class PasefHooks:

    def __init__(self, method_folder):
        logging.info("PasefHooks.__init__({0})".format(method_folder))
        self.method_folder = method_folder
        self.process_cnt = 0
        self.analysis_directory = None

        self.num_of_spectra_sent = 0
        self.num_of_spectra_itr = 0

        self.ms2_frame_counter = 0
        self.ms1_frame_counter = 0
        self.ms1_frame_list = []
        self.ms1_frame_list_size = 2
        self.current_ms1_frame_iteration = 1
        self.flag_for_ms1_frame_list = False
        self.peptide_itr = 1

        # Set up PASER
        self.paser = PASER("192.168.1.176:9092","C:\BrukerEnginePython")
        self.paser.initialize_gpu_workspaces()

    def __str__(self):
        return "Example PASEF hooks for handling MS/MS spectra"

    def leave(self):
        logging.info("PasefHooks.leave()")

    def analysisStarted(self, analysis_directory):
        logging.info("PasefHooks.analysisStarted({0})".format(analysis_directory))
        sqlite_fn = os.path.join(analysis_directory, 'online-msms-spectra.sqlite')

        if os.path.exists(sqlite_fn):
            os.remove(sqlite_fn)

        self.analysis_directory = analysis_directory

    def analysisStopped(self):
        logging.info("PasefHooks.analysisStopped - uninitialize kafka Producer")
        self.paser.uninitialize_gpu_workspaces()
        self.paser.display_duty_cycle(self.analysis_directory, False)
        self.analysis_directory = None

        logging.info("Number of spectra sent: {}".format(self.num_of_spectra_sent))
        logging.info("Number of spectra itr: {}".format(self.num_of_spectra_itr))



    """
    TODO: Seperate functions in different loops as to not slow down duty cycle: ie we can terminate custom
          Proccessing if it takes too long
    PipeLine:
        1) Receive precursor Candidate in processCandidates()
        2) Save information to np array to send through kafka to PASER
        3) Save information to precursor list for sequential frame precursor identification
        4) Identify if precursor was previously searched and if that search was valid / invalid
        
        Create precursor dict which can be reference from engine id
    """
    def processCandidates(self, candidates, analysis_time, make_candidate, enable_individual_collision_energies):
        logging.info("processCandidates")

        min_id = None
        max_id = None
        numOfPrecursors = len(candidates)

        self.ms1_frame_counter += 1
        logging.info("Ms1 Frame: {}, len: {}".format(self.ms1_frame_counter, numOfPrecursors))

        # engine_id is a unique id for precursors that can be used for identification in newMsMsSpectra()
        if numOfPrecursors > 0:
            ids = [c.precursor.engine_id for c in candidates]
            min_id = min(ids)
            max_id = max(ids)

        logging.info('processCandidates(): min/max precursor engine_id = {0}/{1}'.format(min_id, max_id))
        #logging.info("precursor_search_results: {}".format(len(self.paser.precursor_search_results)))

        if len(self.ms1_frame_list) >= self.ms1_frame_list_size:
            self.ms1_frame_list.pop(0)
        ms1_frame_dict = {'frame_min_id': min_id, 'frame_max_id': max_id, 'frame_number': self.ms1_frame_counter, 'prec_dict': {}}
        self.ms1_frame_list.append(ms1_frame_dict)
        logging.info("Current MS1 Frame: {}, Thread_Ms1_Frame: {}".format(self.ms1_frame_counter, self.paser.current_thread_ms1_frame))
        """
        itr = 0
        ms1_frame = np.empty([numOfPrecursors, 13])
        sequential_test_dict = {}
        for c in candidates:

            # Store Precursor info for later use
            ms1_frame[itr][0] = float(0 if c.average_mz is None else c.average_mz)
            ms1_frame[itr][1] = float(0 if c.charge is None else c.charge)
            ms1_frame[itr][2] = float(0 if c.collision_energy is None else c.collision_energy)
            ms1_frame[itr][3] = float(0 if c.desired_num_repetitions is None else c.desired_num_repetitions)
            ms1_frame[itr][4] = float(0 if c.dynamically_excluded is None else c.dynamically_excluded)
            ms1_frame[itr][5] = float(0 if c.intensity is None else c.intensity)
            ms1_frame[itr][6] = float(0 if c.isolation_mz is None else c.isolation_mz)
            ms1_frame[itr][7] = float(0 if c.isolation_width is None else c.isolation_width)
            ms1_frame[itr][8] = float(0 if c.largest_peak_mz is None else c.largest_peak_mz)
            ms1_frame[itr][9] = float(0 if c.monoisotopic_mz is None else c.monoisotopic_mz)
            ms1_frame[itr][10] = float(0 if c.one_over_k0 is None else c.one_over_k0)
            ms1_frame[itr][11] = float(0 if c.scan_number is None else c.scan_number)
            ms1_frame[itr][12] = float(0 if c.scheduling_priority is None else c.scheduling_priority)
            itr += 1

            charge = int(0 if c.charge is None else c.charge)
            monoisotopic_mz = float(0 if c.monoisotopic_mz is None else c.monoisotopic_mz)
            intensity = float(0 if c.intensity is None else c.intensity)
            scan_number = float(0 if c.scan_number is None else c.scan_number)

            # search result
            search_result = None
            if c.monoisotopic_mz:
                bin = math.floor(monoisotopic_mz*100)
                bin_scan = math.floor(scan_number / 10)

                #logging.info("bin: {}, scan bin: {}, charge: {}".format(bin, bin_scan, charge))

                if self.paser.precursor_search_results.get(bin) and \
                   self.paser.precursor_search_results[bin].get(bin_scan) and \
                   self.paser.precursor_search_results[bin][bin_scan].get(charge):

                    # copy result to avoid thread updating. May not be needed. Further testing required.
                    search_result = self.paser.precursor_search_results[bin][bin_scan][charge].copy()
                    #logging.info(search_result)
                    

            # check for sequential frames

            if c.monoisotopic_mz and intensity > 0:
                result = self.paser.search_precursor(monoisotopic_mz, scan_number, intensity, charge)
                if sequential_test_dict.get(result):
                    sequential_test_dict[result] += 1
                else:
                    sequential_test_dict[result] = 1
        

        # Store this frames precursor info

        msg_key = "{} {}".format(self.paser.unique_id, self.ms1_frame_counter)
        self.paser.publish_precursor_spectra(msg_key, ms1_frame.tobytes())

        # Get List of precursors found / not found
        self.paser.add_precursors(ms1_frame)

        logging.info("sequential_test_dict: {}".format(sequential_test_dict))
        logging.info("Number of Candidates: {}".format(len(candidates)))
        """

    def newMsMsSpectra(self, analysis_time, spectra):
        logging.info("PasefHooks.newMsMsSpectra(N={0}, t={1})".format(len(spectra), analysis_time))

        if self.analysis_directory is None:
            return

        self.ms2_frame_counter += 1
        spectra_ms1_frame_number = None     # represents itr of ms1 (1,2,3 ... n)
        prec_dict = None

        for spectrum in spectra:
            self.num_of_spectra_itr += 1

            self.peptide_itr += 1

            # precursor information
            p = spectrum.precursor
            #logging.info("engine_id: {}".format(p.engine_id))
            engine_id = p.engine_id
            charge = 0.0 if p.charge == None else float(p.charge)
            scan_number = float(p.scan_number)
            monoisotopic_mz = float(p.average_mz if p.monoisotopic_mz == None else p.monoisotopic_mz)
            precursor_intensity = float(p.intensity)

            if not spectra_ms1_frame_number:
                logging.info("id: {}, 0 min/max: [{}, {}]".format(engine_id, self.ms1_frame_list[0]['frame_min_id'],  self.ms1_frame_list[0]['frame_max_id']))

                if engine_id >= self.ms1_frame_list[0]['frame_min_id'] and engine_id <= self.ms1_frame_list[0]['frame_max_id']:
                    spectra_ms1_frame_number = self.ms1_frame_list[0]['frame_number']
                    self.flag_for_ms1_frame_list = True
                    prec_dict = self.ms1_frame_list[0]['prec_dict']
                else:
                    if self.flag_for_ms1_frame_list:
                        self.flag_for_ms1_frame_list = False
                        self.current_ms1_frame_iteration = self.ms2_frame_counter
                        self.ms2_frame_counter += 1
                        self.peptide_itr += 1

                    spectra_ms1_frame_number = self.ms1_frame_list[1]['frame_number']
                    prec_dict = self.ms1_frame_list[1]['prec_dict']
                    logging.info("2nd MS1 frame list")

                logging.info(
                    "spectra_ms1_frame_number: {}, ms2_frame_counter: {}, current_ms1_frame_iteration: {}, peptide_itr: {}".format(
                        self.ms2_frame_counter, self.ms2_frame_counter, self.current_ms1_frame_iteration, self.peptide_itr))

            prec_dict[engine_id] = "Test - will me spectra to sum"

            # Skip charge state of 0
            if charge == 0:
                continue

            # quickly parse spectra data
            precursor_information = np.array(
                [scan_number, charge, float(analysis_time), monoisotopic_mz, precursor_intensity,
                 spectra_ms1_frame_number, self.peptide_itr, self.current_ms1_frame_iteration])

            tmp = np.vstack((spectrum.area_values, spectrum.mz_values)).reshape((-1,), order='F')
            spectra_data_and_precursor = np.hstack((precursor_information, tmp))

            msg_key = "{} {}".format(self.paser.unique_id, self.paser.msms_message_counter)
            self.paser.publish_msms_spectra(msg_key, spectra_data_and_precursor.tobytes())
            self.num_of_spectra_sent += 1


    def precursorsExpired(self, max_expired_precursor_id):
        logging.info("PasefHooks.precursorsExpired({0})".format(max_expired_precursor_id))


"""
    PASER (Parrallel database Search Engine in Realtime).
    Utilizes prolucid's GPU workspaces to search fragmented peptide spectra in real time.

    TODO: Figure out how to terminate PASER if they do not start aquisition -> solution On heartbeat failure of init thread
    TODO: Add Seek behavior to look for key offset
    
    TODO: mass tolerance: 10 - 15 ppm
          mobility: 25??
"""

class PASER:
    INSTRUMENT_SPECIFIC_ID = "Ty_Desktop"
    MAX_BYTE_FETCH_INIT = 1024 * 1024 * 10
    REQUEST_TIMEOUT_MS_INIT = 1000
    CONSUMER_TIMEOUT_MS_INIT = 10000
    FETCH_MAX_WAIT_MS_INIT = 100
    INIT_START_CODE = "Start"
    INIT_END_CODE = "End"

    CONSUMER_TIMEOUT_MS_PREC = 100
    FETCH_MAX_WAIT_MS_PREC = 50

    KAFKA_INITIALIZATION_TOPIC = "Initialization"
    KAFKA_SPECTRA_TOPIC = "Spectra"
    KAFKA_PRECURSORS_TOPIC = "Precursor"

    NUM_PREV_PRECURSOR_TO_STORE = 3
    PRECURSOR_MASS_ROUNDING_VALUE = 2

    def __init__(self, ip_address, ip2_gui_path):

        self.ip2_gui_path = ip2_gui_path
        self.ip2_gui_invoke_script = "java -jar ip2gui.jar"
        self.search_info_file = ip2_gui_path + "\\search-id.txt"

        self.unique_id = self.getUniqueId()
        self.kafka_server_ip = ip_address
        self.path_to_ptx_file = ""
        self.path_to_searchxml_file = ""
        self.path_to_fafsta_file = None

        # Tracker for Bytes
        self.totalBytesSent = 0
        self.totalMessagesSent = 0
        self.startTime = int(round(time.time()))
        self.msms_messages_sent = 0
        self.msms_message_counter = 0
        self.precursor_messages_sent = 0


        self.kafka_producer = self.connect_kafka_producer()

        # Precursor consumer Thread
        self.consumer_stop = threading.Event()
        self.precursor_search_results = {}
        self.current_thread_ms1_frame = [-1]
        """
        self.precursor_thread = self.Consumer(args=(self.precursor_search_results,
                                                    self.consumer_stop,
                                                    self.unique_id,
                                                    self.current_thread_ms1_frame))
        self.precursor_thread.start()
        """
        # Precursor selection
        self.list_of_previous_precursors = []
        self.number_of_frames_to_track = 3

    # send initialization message to kafka server to set up GPU workspaces
    def initialize_gpu_workspaces(self):
        # TODO: update kafka side to handle this new input below
        # init_msg = self.create_init_msg()
        # msg = "{} {} {} {}".format(self.unique_id, self.path_to_ptx_file,
        #                                 self.path_to_searchxml_file, self.path_to_fafsta_file)
        self.ip2_gui_invoke()
        msg = "{} {}".format(self.unique_id, self.read_search_id())
        logging.info(msg)
        self.publish_init_message(self.INIT_START_CODE, bytes(msg, encoding='utf-8'))
        self.kafka_producer.flush()

        # verify if initialization of GPU workspaces is complete
        # will provide prompt to user that aquisition is ready
        self.check_initialization_complete()

    def ip2_gui_invoke(self):
        process = subprocess.Popen(self.ip2_gui_invoke_script, shell=True, stdout=subprocess.PIPE,
                                   cwd=self.ip2_gui_path)
        stdout, stderr = process.communicate()
        logging.info(process.returncode)  # is 0 if success


    def read_search_id(self):
        search_file = open(self.search_info_file, "r")
        search_id = search_file.readline()
        logging.info(search_id)
        return search_id;


    def add_precursors(self, ms1_frame):

        table_of_precursors = {}

        for i in range(len(ms1_frame)):
            mz = ms1_frame[i][9]
            sn = ms1_frame[i][11]
            it = ms1_frame[i][5]
            charge = ms1_frame[i][1]

            precursor = {}
            precursor['mz'] = mz
            precursor['charge'] = charge
            precursor['intensity'] = it
            precursor['scan_number'] = sn

            mz_bin = mz * 100
            ppm_offset = mz_bin * 0.00004
            min_bin = math.floor(mz_bin - ppm_offset)
            max_bin = math.floor(mz_bin + ppm_offset) + 1

            min_bin_scan_number = math.floor((sn - 20) / 10)
            max_bin_scan_number = math.floor((sn + 20) / 10) + 1
            #logging.info("{}, {}".format(min_bin, max_bin))
            for i in range(min_bin, max_bin):

                # if scan_number Dict doesnt exist for this bin the create it
                if not table_of_precursors.get(i):
                    table_of_precursors[i] = {}

                scan_number_dict = table_of_precursors[i]

                # Loop over scan number bins
                for j in range(min_bin_scan_number, max_bin_scan_number):

                    # if charge dict doesnt exist then create it
                    if not scan_number_dict.get(j):
                        scan_number_dict[j] = {}

                    charge_dict = scan_number_dict[j]

                    # precursor exists
                    if charge_dict.get(charge):
                        charge_dict[charge].append(precursor)
                    else:
                        charge_dict[charge] = []
                        charge_dict[charge].append(precursor)

        self.list_of_previous_precursors.append(table_of_precursors)

        if len(self.list_of_previous_precursors) > self.number_of_frames_to_track:
            self.list_of_previous_precursors.pop(0)

    def search_precursor(self, mz, sn, it, charge):

        if len(self.list_of_previous_precursors) < self.number_of_frames_to_track:
            return 0

        logging.info("Search:    mz: {}, sn: {}, int: {}, charge: {}".format(mz, sn, it, charge))

        bin = math.floor(float(mz) * 100)
        bin_scan = math.floor(float(sn) / 10)

        number_of_matches = 0

        # loop over previous ms1_frames
        for i in range(len(self.list_of_previous_precursors)):
            precursor_search_results = self.list_of_previous_precursors[i]

            # Verify that precursor exists
            if precursor_search_results.get(bin) and \
                    precursor_search_results[bin].get(bin_scan) and \
                    precursor_search_results[bin][bin_scan].get(charge):

                result = precursor_search_results[bin][bin_scan][charge]
                logging.info("frame#: -{},   {} ".format(len(self.list_of_previous_precursors)-i, str(result)))
                number_of_matches += 1
        return number_of_matches

    # send uninitialization message to kafka server to free up GPU workspaces
    def uninitialize_gpu_workspaces(self):
        self.consumer_stop.set()

        logging.info("uninitialize_gpu_workspaces")
        total_time_elap = int(round(time.time()) - self.startTime)
        msg = "{} {} {} {} {}".format(self.unique_id, self.totalBytesSent,
                                      self.totalMessagesSent, total_time_elap,
                                      self.msms_message_counter)
        self.publish_init_message(self.INIT_END_CODE, bytes(msg, encoding='utf-8'))

        # terminate kafka producer connection
        if self.kafka_producer is not None:
            self.kafka_producer.close()

    # create Kafka producer connection to server
    def connect_kafka_producer(self):
        _producer = None
        try:
            _producer = KafkaProducer(bootstrap_servers=[self.kafka_server_ip], api_version=(0, 10),
                                      batch_size=16384 * 4, linger_ms=50)
        except Exception as ex:
            logging.info('Exception while connecting Kafka Producer')
            logging.info(str(ex))
        finally:
            return _producer

    # send msms spectra
    def publish_msms_spectra(self, key, value):
        self.msms_message_counter += 1
        self.publish_message(self.KAFKA_SPECTRA_TOPIC, key, value)

    # send precursor spectra
    def publish_precursor_spectra(self, key, value):
        self.precursor_messages_sent += 1
        self.publish_message(self.KAFKA_PRECURSORS_TOPIC, key, value)

    # send initialization/de-initialization messages
    def publish_init_message(self, key, value):
        self.publish_message(self.KAFKA_INITIALIZATION_TOPIC, key, value)

    # send message to kafka server
    def publish_message(self, topic_name, key, value):
        try:
            key_bytes = bytes(key, encoding='utf-8')
            value_bytes = value

            self.kafka_producer.send(topic_name, key=key_bytes, value=value_bytes)

            # track bytes/messages sent
            self.totalBytesSent += sys.getsizeof(key_bytes)
            self.totalBytesSent += sys.getsizeof(value)
            self.totalMessagesSent += 1

        except Exception as ex:
            logging.info('Exception in publishing message')
            logging.info(str(ex))

    # Consumes messages from the Initialization topic, and looks for when the initialization is complete
    def check_initialization_complete(self):
        logging.info("initialization consumer thread started")

        consumer = KafkaConsumer(self.KAFKA_INITIALIZATION_TOPIC,
                                 group_id="{} {}".format(self.unique_id, "init"),
                                 bootstrap_servers=[self.kafka_server_ip],
                                 consumer_timeout_ms=self.CONSUMER_TIMEOUT_MS_INIT,
                                 fetch_max_wait_ms=self.FETCH_MAX_WAIT_MS_INIT)

        logging.info("begin message loop")

        for message in consumer:
            # message value and key are raw bytes -- decode if necessary!
            # e.g., for unicode: `message.value.decode('utf-8')`
            logging.info("message (init): %s:%d:%d: key=%s value=%s" % (
                message.topic, message.partition, message.offset, message.key, message.value))
            if (message.key.decode("utf-8") == self.unique_id):
                logging.info("Initialization complete: ready for quisition")
                #self.create_popup_message("Proceed with Aquisition.", "Init completed!")
                x = threading.Thread(target=self.create_popup_message, args=("Proceed with Aquisition.", "Init completed!"))
                x.start()
                return

        logging.info("No inti respose from kafka.")
        #self.create_popup_message("No inti respose from kafka.", "Init failed!")

        x = threading.Thread(target=self.create_popup_message, args=("No inti respose from kafka!", "Init failed!"))
        x.start()

    # Creates a popup message box to notify user of successful/unsuccessful initialization of GPU Workspaces
    def create_popup_message(self, msg, title):
        try:
            ctypes.windll.user32.MessageBoxW(0, msg, title, 1)
        except Exception as ex:
            print(ex)

    # Print total bytes/messages to log file, as well as total runtime
    def print_run_statistics(self):
        logging.info("--------------------------- RealTime Stats ---------------------------")
        logging.info("Total bytes sent: {}".format(self.totalBytesSent))
        logging.info("Total messages sent: {}".format(self.totalMessagesSent))
        total_time_elap = int(round(time.time()) - self.startTime)
        logging.info("Total Time: {}".format(total_time_elap))
        logging.info("Num bytes / s: {}".format(self.totalBytesSent / total_time_elap))
        logging.info("--------------------------- RealTime Stats ---------------------------")

        # create initialization string with unique_id, search.xml, ...

    def create_init_msg(self):

        # Load whole search.xml file into string
        search_xml = open("C:\\Users\\Ty\\Documents\\search.xml")
        xml_file_as_string = search_xml.read()
        search_xml.close()

        # Generate a json object for initializing PASER 
        data = {}
        data['xml'] = xml_file_as_string
        data['unique_id'] = self.unique_id
        data['ptx'] = self.path_to_searchxml_file

        json_data = json.dumps(data)
        return json_data

    # generate unique id for instrument: instrument_ID + date_time stamp
    def getUniqueId(self):
        now = datetime.now()
        current_time = now.strftime("%H_%M_%S")
        return "{}{}".format(self.INSTRUMENT_SPECIFIC_ID, current_time)

    def display_duty_cycle(self, analysis_dir, plot):
        frame_id_low = 1
        frame_id_high = 0

        conn = sqlite3.connect(os.path.join(analysis_dir, "analysis.tdf"))

        # check if range for frame ids where specified
        if frame_id_high == 0:
            frame_id_high = conn.execute("SELECT MAX(Id) from Frames").fetchone()[0]

        # Plot MS1 TIC
        tmp = conn.execute(
            "SELECT Id, SummedIntensities FROM Frames WHERE MsMsType=0 AND Id BETWEEN {0} AND {1} ORDER BY Id".format(
                frame_id_low, frame_id_high)).fetchall()
        tic_ids = np.array([tuple[0] for tuple in tmp])
        tic_intensities = np.array([tuple[1] for tuple in tmp])
        # plt.plot(tic_ids, tic_intensities)
        # plt.xlabel('frame number')
        # plt.ylabel('MS1 TIC')

        # Get times for precursor selection + scheduling (scheduling is really small) and frame-acquisition times
        tmp = conn.execute(
            "SELECT f.Id, p.Value FROM Frames f JOIN Properties p ON p.Frame=f.Id AND p.Property=(SELECT Id FROM PropertyDefinitions WHERE PermanentName='PrecSel_CompleteTime') AND p.Value NOT NULL ORDER BY f.Id").fetchall()
        precsel_ids = [tuple[0] for tuple in tmp]
        precsel_times = [tuple[1] for tuple in tmp]
        tmp = conn.execute("SELECT Id, Time FROM Frames WHERE Id BETWEEN {0} AND {1} ORDER BY Id".format(frame_id_low,
                                                                                                         frame_id_high)).fetchall()
        ids = np.array([tuple[0] for tuple in tmp])
        times = np.array([tuple[1] for tuple in tmp])
        timediffs = times[1:] - times[0:-1]

        # Get frame-submission times
        tmp = conn.execute(
            "SELECT f.Id, p.Value FROM Frames f JOIN Properties p ON p.Frame=f.Id AND p.Property=(SELECT Id FROM PropertyDefinitions WHERE PermanentName='Timing_SubmitFrame') AND p.Value NOT NULL ORDER BY f.Id").fetchall()
        submit_ids = [tuple[0] for tuple in tmp]
        submit_times = [tuple[1] for tuple in tmp]

        # Get theoretical time per frame [us]
        def get_unique_value(query):
            tmp = conn.execute(query).fetchall()
            if len(tmp) != 1:
                raise RuntimeError('expect exactly one result row')
            return tmp[0][0]

        cycletime_sec = 1e-6 * get_unique_value(
            "SELECT DISTINCT(p.Value) FROM Properties p WHERE p.Property = (SELECT Id FROM PropertyDefinitions WHERE PermanentName='Digitizer_ExtractTriggerTime')")
        # print cycletime_sec
        numscans = get_unique_value("SELECT DISTINCT(NumScans) FROM Frames")
        # print numscans
        quenchtime_sec = 1e-3 * get_unique_value(
            "SELECT DISTINCT(p.Value) FROM Properties p WHERE p.Property = (SELECT Id FROM PropertyDefinitions WHERE PermanentName='Collision_QuenchTime_Set')")
        # print quenchtime_sec
        exp_frame_time = numscans * cycletime_sec + quenchtime_sec

        # number of empty MS frames
        empty_ms = get_unique_value("SELECT COUNT(*) FROM Frames WHERE NumPeaks=0 AND MsMsType=0")
        empty_msms = get_unique_value("SELECT COUNT(*) FROM Frames WHERE NumPeaks=0 AND MsMsType=8")

        # print exp_frame_time
        logging.info("Number of empty MS frames {}".format(empty_ms))
        logging.info("Number of empty MSMS frames {}".format(empty_msms))
        logging.info("Average abs(time excess) = {0:.2f} %".format(
            100 * np.mean(np.abs(timediffs - exp_frame_time)) / exp_frame_time))
        logging.info(
            "Average time excess = {0:.2f} %".format(100 * np.mean(timediffs - exp_frame_time) / exp_frame_time))
        logging.info("Abs deviation from expected time {0:.6f}s".format(np.mean(timediffs - exp_frame_time)))
        if 1 < len(precsel_times):
            logging.info("Average time precursor search + scheduling: {0:.3f}s".format((np.mean(precsel_times))))
        logging.info("expected time for frame: {0}s".format(exp_frame_time))
        logging.info("number of scans: {0}".format(numscans))
        logging.info("trigger period: {0}".format(cycletime_sec))
        logging.info("quench time: {0}".format(quenchtime_sec))

        # Plot results
        if plot:
            plt.figure()
            plt.plot(ids[0:-1], timediffs)

            plt.plot(precsel_ids, precsel_times, 'o', alpha=0.1)
            plt.plot(submit_ids, submit_times, 'x', alpha=0.1)
            plt.plot([ids[0], ids[-1]], [exp_frame_time, exp_frame_time], color=[1, 0, 0])
            plt.legend(('time delta between consecutive frames', 'time for precsel + scheduling'))
            plt.ylabel('time / sec')
            plt.xlabel('frame number')
            plt.ylim([0, 0.6])
            plt.show()

    # Precursor Consumer Thread
    # TODO: Hashtable for scan_number
    # TODO: add counter for number of scans
    class Consumer(threading.Thread):

        def __init__(self, group=None, target=None, name=None,
                     args=(), kwargs=None):
            threading.Thread.__init__(self, group=group, target=target, name=name)
            self.args = args
            self.precursor_dict = self.args[0]
            self.consumer_stop = self.args[1]
            self.id = self.args[2]
            self.current_frame = self.args[3]

            self.mz_bin_multiplier = 100
            self.ppm_decimal_value = 0.00004
            self.scan_number_bin_divisor = 10
            self.scan_number_threshold = 10


            self.consumer = KafkaConsumer(group_id=str("precursor-consumer-thread"),
                                          bootstrap_servers="172.29.80.71:9092",
                                          auto_offset_reset='latest')
            self.consumer.subscribe(['SearchedPrecursors'])

            return

        def run(self):

            logging.info("Run")

            precursorDict = {}
            try:

                for message in self.consumer:
                    key = message.key.decode('utf8')
                    value = message.value.decode('utf8')
                    values = value.split(", ")

                    if len(values) is not 6:
                        continue

                    if key != self.id:
                        logging.info("skipping old records, Old {}, New {}".format(key, self.id))
                        continue

                    precursor = {}
                    precursor['valid'] = False if float(values[3]) == 0.0 else True
                    precursor['mz'] = float(values[0])
                    precursor['charge'] = int(values[1])
                    precursor['intensity'] = float(values[2])
                    precursor['theo_mz'] = float(values[3])
                    precursor['ms1FrameNumber'] = int(values[4])
                    precursor['scan_number'] = float(values[5])
                    precursor['times_searched'] = 1
                    precursor['Sequence'] = "abcd"

                    precursor['min_mz_bin'] = -1
                    precursor['max_mz_bin'] = -1
                    precursor['min_sn_bin'] = -1
                    precursor['max_sn_bin'] = -1


                    if self.current_frame[0] != precursor['ms1FrameNumber']:
                        self.current_frame[0] = precursor['ms1FrameNumber']

                    # Check if Precursor exists in list
                    mz_bin_value = math.floor(precursor['mz'] * self.mz_bin_multiplier)
                    scan_bin_value = math.floor(precursor['scan_number']/self.scan_number_bin_divisor)
                    charge_bin_value = precursor['charge']
                    mz_ppm = precursor['mz'] * self.ppm_decimal_value

                    if self.precursor_dict.get(mz_bin_value) and \
                            self.precursor_dict[mz_bin_value].get(scan_bin_value) and \
                            self.precursor_dict[mz_bin_value][scan_bin_value].get(charge_bin_value):

                        # Precursor exists: verify if it is the same precursor
                        precursor_result_list = self.precursor_dict[mz_bin_value][scan_bin_value][charge_bin_value]
                        matched_precursor_from_list = None
                        number_of_macthes = 0
                        for p in precursor_result_list:

                            if abs(p['mz'] - precursor['mz']) < mz_ppm and \
                                    abs(p['scan_number'] - precursor['scan_number']) < self.scan_number_threshold and \
                                    p['charge'] == precursor['charge']:

                                number_of_macthes += 1

                                # check if matched precursor already selected: then update for whichever is closer
                                if matched_precursor_from_list is None:
                                    matched_precursor_from_list = p
                                else:
                                    mz_difference_matched_prec = abs(precursor['mz'] - matched_precursor_from_list['mz'])
                                    mz_difference_curr_prec = abs(precursor['mz'] - p['mz'])

                                    sn_difference_matched_prec = abs(precursor['scan_number'] - matched_precursor_from_list['scan_number'])
                                    sn_difference_curr_prec = abs(precursor['scan_number'] - p['scan_number'])

                                    # Normalize the differences for distance equation
                                    matched_prec_normalized_mz_difference = mz_difference_matched_prec / mz_ppm
                                    curr_prec_normalized_mz_difference = mz_difference_curr_prec / mz_ppm
                                    matched_prec_normalized_sn_difference = sn_difference_matched_prec / self.scan_number_threshold
                                    curr_prec_normalized_sn_difference = sn_difference_curr_prec / self.scan_number_threshold

                                    # calculate the Euclidean distance
                                    matched_prec_normalized_difference = matched_prec_normalized_mz_difference**2 + \
                                                                        matched_prec_normalized_sn_difference**2

                                    curr_prec_normalized_difference = curr_prec_normalized_mz_difference**2 + \
                                                                        curr_prec_normalized_sn_difference**2

                                    # curr_precursor is closer to precursor than matched_precursor
                                    if curr_prec_normalized_difference < matched_prec_normalized_difference:
                                        matched_precursor_from_list = p
                                        logging.info("Precursor is closer! {}, {}".format(curr_prec_normalized_difference, matched_prec_normalized_difference))

                        # update matched precursor with new information
                        if matched_precursor_from_list is not None:

                            average_mz = (matched_precursor_from_list['mz'] + precursor['mz'])/2
                            average_sn = (matched_precursor_from_list['scan_number'] + precursor['scan_number'])/2
                            old_mz = matched_precursor_from_list['mz']
                            old_sn = matched_precursor_from_list['scan_number']
                            times_searched = matched_precursor_from_list['times_searched'] + 1

                            # mz and scan bins updated
                            update_bins = math.floor(old_mz*100) != math.floor(average_mz*100) or math.floor(old_sn/10) != math.floor(average_sn/10)

                            if update_bins:
                                self.remove_precursor_from_dict(matched_precursor_from_list)

                                precursor['mz'] = average_mz
                                precursor['scan_number'] = average_sn
                                precursor['times_searched'] = times_searched
                                self.add_precursor_to_dict(precursor)
                            else:
                                matched_precursor_from_list['mz'] = average_mz
                                matched_precursor_from_list['scan_number'] = average_sn
                                matched_precursor_from_list['times_searched'] = times_searched


                            logging.info("number of times matched: {}".format(number_of_macthes))


                        # Add precursor to list/dicts
                        else:
                            self.add_precursor_to_dict(precursor)

                    # precursor doesnt exist in list
                    else:
                        self.add_precursor_to_dict(precursor)

                    if self.consumer_stop.is_set():
                        break

            except Exception as ex:
                logging.info('Exception in Thread Precusorsor: \n {} \n {}'.format(ex, ex.with_traceback()))
                logging.info(traceback.format_exc())
                logging.info(sys.exc_info()[2])


            self.consumer.close()

        # returns min/max bin numbers
        def add_precursor_to_dict(self, precursor):
            mz_bin = precursor['mz'] * self.mz_bin_multiplier
            mz_ppm_offset = mz_bin * self.ppm_decimal_value
            min_mz_bin = math.floor(mz_bin - mz_ppm_offset)
            max_mz_bin = math.floor(mz_bin + mz_ppm_offset) + 1

            min_bin_scan_number = math.floor((precursor['scan_number'] - self.scan_number_threshold) / self.scan_number_bin_divisor)
            max_bin_scan_number = math.floor((precursor['scan_number'] + self.scan_number_threshold) / self.scan_number_bin_divisor) + 1

            precursor['min_mz_bin'] = min_mz_bin
            precursor['max_mz_bin'] = max_mz_bin
            precursor['min_sn_bin'] = min_bin_scan_number
            precursor['max_sn_bin'] = max_bin_scan_number

            #logging.info("{}, {}".format(min_mz_bin, max_mz_bin))
            #logging.info("{}, {}".format(min_bin_scan_number, max_bin_scan_number))

            # loop over m/z bins
            for i in range(min_mz_bin, max_mz_bin):

                # if scan_number Dict doesnt exist for this bin the create it
                if not self.precursor_dict.get(i):
                    self.precursor_dict[i] = {}

                scan_number_dict = self.precursor_dict[i]

                # Loop over scan number bins
                for j in range(min_bin_scan_number, max_bin_scan_number):

                    # if charge dict doesnt exist then create it
                    if not scan_number_dict.get(j):
                        scan_number_dict[j] = {}

                    charge_dict = scan_number_dict[j]

                    # precursor exists
                    if charge_dict.get(precursor['charge']):
                        charge_dict[precursor['charge']].append(precursor)
                    else:
                        charge_dict[precursor['charge']] = []
                        charge_dict[precursor['charge']].append(precursor)

        def remove_precursor_from_dict(self, precursor):

            min_mz_bin = precursor['min_mz_bin']
            max_mz_bin = precursor['max_mz_bin']
            min_bin_scan_number = precursor['min_sn_bin']
            max_bin_scan_number = precursor['max_sn_bin']
            charge = precursor['charge']

            #logging.info("{}, {}".format(min_mz_bin, max_mz_bin))
            #logging.info("{}, {}".format(min_bin_scan_number, max_bin_scan_number))

            # loop over m/z bins
            for i in range(min_mz_bin, max_mz_bin):

                scan_number_dict = self.precursor_dict[i]

                # Loop over scan number bins
                for j in range(min_bin_scan_number, max_bin_scan_number):

                    scan_number_dict[j][charge].remove(precursor)
                    #logging.info("removedPrecurosr")

