package com.ip2.paser;

import com.ip2.paser.util.PaserConstants;
import javafx.util.Pair;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.apache.kafka.common.serialization.StringSerializer;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import reactor.core.Disposable;
import reactor.core.publisher.EmitterProcessor;
import reactor.core.publisher.Flux;
import reactor.core.publisher.FluxSink;
import reactor.core.scheduler.Scheduler;
import reactor.core.scheduler.Schedulers;
import reactor.kafka.receiver.KafkaReceiver;
import reactor.kafka.receiver.ReceiverOptions;

import org.json.simple.JSONObject;
import org.json.simple.parser.*;
import reactor.kafka.sender.KafkaSender;
import reactor.kafka.sender.SenderOptions;
import reactor.kafka.sender.SenderRecord;
import reactor.kafka.sender.SenderResult;

import java.io.File;
import java.util.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;


public class InitializationConsumer {

    private Hashtable<String, PaserInstance> passerInstances;
    private PaserConstants paserConstants;
    private Disposable initReceiverDisposable;
    private KafkaSender<String, String> sender;
    private static Logger logger = Logger.getLogger(InitializationConsumer.class);

    public InitializationConsumer() {
        paserConstants = PaserConstants.getInstance();
        passerInstances = new Hashtable<>();
    }

    /**
     * Sets up a reactive Flux object for handling PASER initialization messages. Utilizes an
     * elastic scheduler to create separate threads for each inbound message, and sends a response
     * for both start and end messages upon completion.
     *
     * A filter is used to filter out messages from being sent.
     *
     * @return Flux subscribable object
     */
    private Flux<?> InitFlux() {
        sender = initSender(initSenderOptions());
        EmitterProcessor<Pair<String, String>> processor = EmitterProcessor.create();
        FluxSink<Pair<String, String>> incoming = processor.sink();
        Flux<?> inFlux = KafkaReceiver.create(initReceiverOptions(Collections.singleton(paserConstants.getInitializationTopic())))
                .receiveAutoAck()
                .concatMap(r -> r)
                .doOnError(e -> e.printStackTrace())
                .doOnNext(m -> incoming.next(new Pair(m.key(), m.value())));

        // Elastic Scheduler to handle incoming initialization messages
        Scheduler InitScheduler = Schedulers.newElastic("Initialization Scheduler");
        Flux<SenderResult<Integer>> stream1 = sender.send(processor
                .publishOn(InitScheduler)
                .map(p -> SenderRecord.create(
                        processInitMessage(p), 1))
                .doOnError(e -> e.printStackTrace())
                .filter((p) -> isInitCompleteMessage(p)));


        AtomicReference<Disposable> cancelRef = new AtomicReference<>();
        Consumer<AtomicReference<Disposable>> cancel = cr -> {
            Disposable c = cr.getAndSet(null);
            if (c != null)
                c.dispose();
        };
        return stream1.doOnSubscribe(s -> cancelRef.set(inFlux.subscribe()))
                      .doOnCancel(() -> {
                        cancel.accept(cancelRef);
                        close("doOnCancel");
                      })
                      .doOnError(e -> e.printStackTrace())
                      .doOnTerminate(() -> close("doOnTerminate"));

    }

    private ProducerRecord<String, String> processInitMessage(Pair<String, String> message) {
        logger.info("Message :" + message.toString());
        ProducerRecord record;

        // Filter through initialization message keys
        if(message.getKey().equals(paserConstants.getInitStartKey())) {
            logger.debug("Init Start message (start paser instance): " + message.toString());
            record = startPaserInstance(message.getValue());
        } else if(message.getKey().equals(paserConstants.getInitEndKey())) {
            logger.debug("Init End message (shutdown paser instance): " + message.toString());
            record = endPaserInstance(message.getValue());
        } else if(message.getKey().equals(paserConstants.getInitCompletedKey())) {
            // Do nothing, will be handled by producer
            logger.debug("Init completed message (Do nothing): " + message.toString());
            record = new ProducerRecord<>(paserConstants.getInitializationTopic(), "DontSend", "This message wont be sent");
        }  else {
            logger.error("Incorrect Init Key: " + message.toString());
            record = new ProducerRecord<>(paserConstants.getInitializationTopic(), "DontSend", "This message wont be sent");
        }
        return record;
    }

    private ProducerRecord<String, String> startPaserInstance(String messageValue){
        String uniqueId;
        String searchId;

        // parsing incoming json message"
        try {
            Object obj = new JSONParser().parse(messageValue);
            JSONObject jo = (JSONObject) obj;

            uniqueId = (String) jo.get("id");
            searchId = (String) jo.get("search_id");

            logger.info(" Received uniqueId from Producer " + uniqueId);
            logger.info(" Received searchId from Producer " + searchId);

        } catch (Exception e){
            logger.error("Problem parsing PASER Init message!");
            logger.error(e);

            return new ProducerRecord<>(paserConstants.getInitializationTopic(), paserConstants.getInitFailedKey(), "Init Not Complete: Parse start args failed");
        }

        // Catch duplicate Initializations
        if(passerInstances.containsKey(uniqueId)) {
            logger.error("Already created workspace for this ID : " + uniqueId);
            return new ProducerRecord<>(paserConstants.getInitializationTopic(), paserConstants.getInitFailedKey(), "Init Not Complete: Duplicate Workspace");
        }

        // TODO: This is temporary
        // Loop through PASER instances and remove existing Instances
        for(String key : passerInstances.keySet()){
            logger.info("Removing existing PASER Instance");
            PaserInstance pi = passerInstances.get(key);
            passerInstances.remove(key);
            pi.free();
        }

        PaserInstance passerInstance = new PaserInstance(uniqueId);
        passerInstances.put(uniqueId, passerInstance);
        passerInstance.run();

        return new ProducerRecord<>(paserConstants.getInitializationTopic(), paserConstants.getInitCompletedKey(), uniqueId);
    }

    private ProducerRecord<String, String> endPaserInstance(String messageValue){
        String uniqueId;
        long totalBytesSent;
        long totalMessagesSent;
        long totalTime;
        long lastMessageNum;

        // parsing incoming json message"
        try {
            Object obj = new JSONParser().parse(messageValue);
            JSONObject jo = (JSONObject) obj;

            uniqueId = (String) jo.get("id");
            totalBytesSent = (long) jo.get("totalBytesSent");
            totalMessagesSent = (long) jo.get("totalMessagesSent");
            totalTime = (long) jo.get("totalTime");
            lastMessageNum = (long) jo.get("msmsMessageCounter");

            logger.info("total bytes sent: " + totalBytesSent);
            logger.info("total msg sent: " + totalMessagesSent);
            logger.info("total time: " + totalTime);
            logger.info("lastMessageNum: " + lastMessageNum);

        } catch (Exception e){
            logger.error("Problem parsing PASER uninit message!");
            logger.error(e);
            return new ProducerRecord<>(paserConstants.getInitializationTopic(), "Un-init Failed", "Un init Failed!! Failed to parse end args");
        }

        // Free resources and remove from Paser Instance List
        if(passerInstances.containsKey(uniqueId)) {
            PaserInstance passer = passerInstances.get(uniqueId);
            passerInstances.remove(uniqueId);
            passer.free();
        }

        return new ProducerRecord<>(paserConstants.getInitializationTopic(), "Un-init Complete", "Un init Complete!!");
    }

    private void close(String s) {
        logger.info("Ending PASER");
    }

    public void free(){
        initReceiverDisposable.dispose();
        sender.close();

        for(String key : passerInstances.keySet()){
            logger.info("Removing existing PASER Instance");
            PaserInstance pi = passerInstances.get(key);
            passerInstances.remove(key);
            pi.free();
        }
    }

    private boolean isInitCompleteMessage(SenderRecord sr) {
        if (sr.key() == paserConstants.getInitCompletedKey() || sr.key() == paserConstants.getInitFailedKey()){
            return true;
        }
        return false;
    }

    public void consumeMessages() {
        logger.info("Initialization Consumer Started");
        initReceiverDisposable = InitFlux().subscribe();
    }

    private  ReceiverOptions<String, String> initReceiverOptions() {
        Map<String, Object> props = new HashMap<>();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG,
                paserConstants.getKafkaIP());
        props.put(ConsumerConfig.GROUP_ID_CONFIG,
                "InitializationConsumer");
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG,
                StringDeserializer.class.getName());
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG,
                StringDeserializer.class.getName());
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "latest");
        return ReceiverOptions.create(props);
    }

    private ReceiverOptions<String, String> initReceiverOptions(Collection<String> topic) {
        return initReceiverOptions()
                //.commitInterval(Duration.ofSeconds(10))
                .addAssignListener(partitions -> logger.info("onPartitionsAssigned {}" + partitions))
                .addRevokeListener(partitions -> logger.info("onPartitionsRevoked {}" + partitions))
                .subscription(topic);
    }

    private SenderOptions<String, String> initSenderOptions() {
        Map<String, Object> senderProps = new HashMap<>();
        senderProps.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, paserConstants.getKafkaIP());
        senderProps.put(ProducerConfig.CLIENT_ID_CONFIG, "Initialization-Sender");
        senderProps.put(ProducerConfig.ACKS_CONFIG, "all");
        senderProps.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG,
                StringSerializer.class.getName());
        senderProps.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG,
                StringSerializer.class.getName());
        return SenderOptions.create(senderProps);
    }

    private KafkaSender<String, String> initSender(SenderOptions<String, String> senderOptions) {
        return KafkaSender.create(senderOptions);
    }


    public static void main(String args[]) throws Exception {
        System.out.println("main: RealTimeSpectraAnalysisHandlerReact");
        InitializationConsumer rtsah = new InitializationConsumer();
        rtsah.consumeMessages();
    }
}
