package com.ip2.paser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

import com.ip2.paser.util.PaserConstants;
import edu.scripps.pms.mspid.CudaDeviceContext;
import edu.scripps.pms.mspid.GPUSpectraWorkSpace;
import edu.scripps.pms.mspid.ScoredPeptideHit;
import edu.scripps.pms.mspid.SearchResult;
import edu.scripps.pms.util.spectrum.PeakList;
import edu.scripps.pms.util.spectrum.Zline;
import javafx.util.Pair;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.*;

import org.apache.log4j.Logger;
import reactor.core.Disposable;
import reactor.core.publisher.EmitterProcessor;
import reactor.core.publisher.Flux;
import reactor.core.publisher.FluxSink;
import reactor.core.scheduler.Scheduler;
import reactor.core.scheduler.Schedulers;
import reactor.kafka.receiver.ReceiverOptions;
import reactor.kafka.receiver.KafkaReceiver;
import reactor.kafka.sender.KafkaSender;
import reactor.kafka.sender.SenderOptions;
import reactor.kafka.sender.SenderRecord;
import reactor.kafka.sender.SenderResult;

import static jcuda.driver.JCudaDriver.cuCtxSetCurrent;

public class Ms2SpectraConsumer {

    private final String receiverGroupID;
    private final String senderGroupID;
    private final String outputDir;

    private int number_ofSpectra_received_ms2 = 0;
    private int number_ofSpectra_received_Db = 0;

    // Arguments
    private ProlucidWorkSpaceController prolucidController;
    private GPUSpectraWorkSpace gpuSpectraWorkSpaces[];
    private CudaDeviceContext context;
    private final String uniqueId;
    private final String instrumentSpecificId;
    private int numberOfWorkSpaceThreads;

    // Disposables
    private KafkaSender<String, String> databaseSearchResultSender;
    private Scheduler databaseSeachScheduler;
    private Scheduler ms2FileScheduler2;
    private Disposable kafkaReceivorDisposable;

    // Thread Specific Items
    private Hashtable<String, Integer> uniqueThreadIndexHashTable;
    private BufferedWriter bufWritersSqt[];
    private BufferedWriter bufWriterMs2;
    private MessageTracker msgTrackersSqt[];
    private MessageTracker msgTrackerMs2;

    // Trackers
    private long waitTimeBeforeThreadIsConsideredFinished = 5000;
    private long timeSinceLastSearch;

    private static Logger spectraLogger = Logger.getLogger(Ms2SpectraConsumer.class);
    private Logger paserLogger = Logger.getLogger(PaserInstance.class);
    private PaserConstants paserConstants;

    private boolean use_gpu;


    /**
     * Reactive Kafka Consumer to handle searching spectra and creating sqt files
          * @param id the unique id associated with this connection to the instrument
     */
    public Ms2SpectraConsumer(String pathToPtxFile, String pathToSearchXmlFile, String id) {

        uniqueId = id;
        paserConstants = PaserConstants.getInstance();
        use_gpu = paserConstants.getUseGpu();

        numberOfWorkSpaceThreads = paserConstants.getNumberOfThreadsForGPUSearch();
        uniqueThreadIndexHashTable = new Hashtable<>();
        outputDir = paserConstants.getDataOutPath() + "/" + uniqueId;
        timeSinceLastSearch = System.currentTimeMillis();

        instrumentSpecificId = uniqueId.split("-")[0];
        receiverGroupID = instrumentSpecificId + "Ms2SpectraConsumer";
        senderGroupID = instrumentSpecificId + "Ms2SpectraSender";

        try {
            // Initialize gpu workspaces with passes file
            if (use_gpu){
                prolucidController = new ProlucidWorkSpaceController(pathToPtxFile, pathToSearchXmlFile, paserConstants.getNumberOfThreadsForGPUSearch());
                gpuSpectraWorkSpaces = prolucidController.getGpuSpectraWorkSpaces();
                context = prolucidController.getContext();
            } else {
                prolucidController = null;
                gpuSpectraWorkSpaces = new GPUSpectraWorkSpace[10];
                context = null;

                try
                {
                    Thread.sleep(2000);
                }
                catch (InterruptedException e)
                {
                    Thread.currentThread().interrupt(); // restore interrupted status
                }
            }
        } catch (Exception e){
            spectraLogger.error(e.getMessage());
            e.printStackTrace();
            free();
            return;
        }

        getBufferedWriters();
        getMessageTrackers();
    }

    /**
     * Reactive consumer to handle Spectra messages.
     */
    public void consumeMessages() {
        spectraLogger.info("[Starting Method!! consumeMessages()");
        kafkaReceivorDisposable = flux().subscribe();
        spectraLogger.info("Ending Method!! consumeMessages()");
    }

    //flux.buffer?
    public Flux<?> flux() {
        databaseSearchResultSender = sender(senderOptions());
        EmitterProcessor<Pair<String, byte[]>> processor = EmitterProcessor.create();
        FluxSink<Pair<String, byte[]>> incoming = processor.sink();
        AtomicInteger assigner = new AtomicInteger(0);
        Flux<?> inFlux = KafkaReceiver.create(receiverOptions(Collections.singleton(paserConstants.getMs2SpectraTopic())))
                .receiveAutoAck()
                .concatMap(r -> r)
                .doOnNext(m -> incoming.next(new Pair(m.key(), m.value())));


        // newParallel scheduler to handle DB search
        // TODO: Why use thread-name as name?
        databaseSeachScheduler = Schedulers.newParallel("Database Scheduler", numberOfWorkSpaceThreads, false);
        Flux<SenderResult<Integer>> stream1 = databaseSearchResultSender.send(processor
                .groupBy(m -> assigner.incrementAndGet() % numberOfWorkSpaceThreads)
                .flatMap(partitionFlux -> partitionFlux.publishOn(databaseSeachScheduler)
                        .map(p -> SenderRecord.create(processDbSearch(p), 1))
                        .filter(p -> !isDuplicate(p))));

        // single scheduler to handle writing ms2 files
        ms2FileScheduler2 = Schedulers.newSingle("Ms2File Scheduler", false);
        Flux<String> stream2 = processor.publishOn(ms2FileScheduler2).map(p -> processWriteMs2File(p));

        AtomicReference<Disposable> cancelRef = new AtomicReference<>();
        Consumer<AtomicReference<Disposable>> cancel = cr -> {
            Disposable c = cr.getAndSet(null);
            if (c != null)
                c.dispose();
        };
        return Flux.merge(stream1, stream2)
                .doOnSubscribe(s -> cancelRef.set(inFlux.subscribe()))
                .doOnCancel(() -> {
                    cancel.accept(cancelRef);
                    close("doOnCancel");
                })
                .doOnError(e -> e.printStackTrace())
                .doOnTerminate(() -> close("doOnTerminate"));

    }

    private boolean isDuplicate(SenderRecord sr) {
        if (sr.key() == null){
            return true;
        }
        return false;
    }

    private void close(String s) {
        spectraLogger.info(s);

    }


    public ProducerRecord<String, String> processDbSearch(Pair<String, byte[]> message) {
        //System.out.println("processDbSearch     Thread: " + Thread.currentThread().getName());

        String keySplit[] = message.getKey().split(" ");
        String msgId = keySplit[0];
        int msgNumber = Integer.parseInt(keySplit[1]);

        // Skip old records, Message will not be sent
        if(msgId.compareTo(uniqueId) != 0 ){
            spectraLogger.info("SQT Skipping spectra for another id: " + msgId + ", MsgNum: " + msgNumber);
            return new ProducerRecord<>("Skipping", null, "This message will be discarded");
        }

        ArrayList<SearchResult> listOfResults = new ArrayList<>();

        // if thread name is not in hashtable, then add it, otherwise get unique id from table
        String threadName = Thread.currentThread().getName();
        if(!uniqueThreadIndexHashTable.containsKey(threadName)) {
            if (use_gpu) cuCtxSetCurrent(context.getContext());
            setUpThreadForGpuSearch(threadName);
        }

        // threadIndex = unique id for each thread under scheduler. Used to access objects in hash tables
        int threadIndex = uniqueThreadIndexHashTable.get(threadName);
        msgTrackersSqt[threadIndex].add(message, msgNumber);
        number_ofSpectra_received_Db++;
        if(msgNumber % 100 == 0) spectraLogger.info("[Db Search] msgNumber: " + msgNumber +
                " number_ofSpectra_received_Db: " + number_ofSpectra_received_Db);

        // Parse the message
        SpectraScanParser scan = new SpectraScanParser(message.getValue());
        PeakList peakList = scan.getPeakList();
        Iterator<Zline> zItr = peakList.getZlines();
        while(zItr.hasNext())
        {
            Zline zline = zItr.next();
            /*
            if(zline.getM2z() >= 6000 || zline.getM2z() <= 600){
                spectraLogger.info("Mass > 6000: " + msgId + ", MsgNum: " + msgNumber);
                return new ProducerRecord<>("Skipping", null, "Mass > 6000");
            }
            */

            SearchResult sr;
            try{
                if (use_gpu){
                    sr = gpuSpectraWorkSpaces[threadIndex].search(zline,peakList);
                } else {
                    spectraLogger.info("skipping");
                    sr = null;
                    String output = 0 + ", " + 0 + ", " + 0 + ", " + 0 + ", " + 0 + ", " + msgNumber;
                    return new ProducerRecord<>("SearchedPrecursors", this.uniqueId, output);
                }
            } catch (Exception e){
                spectraLogger.error(scan.debugString());
                spectraLogger.error(e.getMessage());
                return new ProducerRecord<>("Error", null, "Error: " + e);
            }

            if(sr == null){
                return new ProducerRecord<>("Skipping", null, "oob");
            }

            listOfResults.add(sr);

            // write output to filename[thread number].sqt file
            BufferedWriter bw = bufWritersSqt[uniqueThreadIndexHashTable.get(threadName)];
            try{
                bw.write(sr.outputResults());
            } catch (IOException e) {
                spectraLogger.error("An error occurred writing to file.");
                spectraLogger.error(e);
                spectraLogger.error(e.getStackTrace());
            }
        }

        //Precursor Identification, Information gets used by the producer to select future precursors
        Double theoreticalMass = getPrecursorInformation(listOfResults);
        String output = scan.precursorMass + ", " + scan.charge + ", " + scan.intensity + ", " + theoreticalMass + ", " + scan.ms1FrameNumber + ", " + msgNumber;

        timeSinceLastSearch = System.currentTimeMillis();
        return new ProducerRecord<>("SearchedPrecursors", this.uniqueId, output);
    }

    private Double getPrecursorInformation(ArrayList<SearchResult> listOfResults) {
        float precursorScore = 0;
        float precursorMz;
        int precursorChargeState;
        float precursork0;
        float precursorRet;

        if(listOfResults == null) return 0.0;

        for (SearchResult sr : listOfResults){
            if (sr == null || sr.getFinalResultList() == null) return 0.0;
            Iterator<ScoredPeptideHit> itr = sr.getFinalResultList().iterator();

            while(itr.hasNext()){
                ScoredPeptideHit p = itr.next();

                //TODO: Figure out the correct value to use here -> Xcorr
                if(p.getPScore() <= 0.01){ //FDR of 1%
                    return p.getTheorMass();
                }

                if(false) {
                    spectraLogger.info(p.getPrimaryScore());
                    spectraLogger.info(p.getPrimaryRank());
                    spectraLogger.info(p.getXCorr());
                    spectraLogger.info(p.getXCorrRank());
                    spectraLogger.info(p.getPScore());
                    spectraLogger.info(p.getPScoreRank());
                    spectraLogger.info(p.getSecondaryScore());
                    spectraLogger.info(p.getSecondaryRank());
                }
            }
        }
        return 0.0;

    }


    public String processWriteMs2File(Pair<String, byte[]> message) {
        //System.out.println("processWriteMs2File     Thread: " + Thread.currentThread().getName());

        String keySplit[] = message.getKey().split(" ");
        String msgId = keySplit[0];
        int msgNumber = Integer.parseInt(keySplit[1]);

        // Skip messages from different instrument
        if(msgId.compareTo(uniqueId) != 0 ){
            spectraLogger.info("Ms2 Skipping spectra for another id: " + msgId + ", MsgNum: " + msgNumber);
            return "Skipping";
        }

        number_ofSpectra_received_ms2++;
        msgTrackerMs2.add(message, msgNumber);
        if(number_ofSpectra_received_ms2 % 100 == 0) spectraLogger.info("[Ms2 File] number_ofSpectra_received: " + number_ofSpectra_received_ms2);

        // Parse the incoming message
        SpectraScanParser parse = new SpectraScanParser(message.getValue());
        String ms2Data = parse.Ms2ToString();

        try {
            bufWriterMs2.write(ms2Data);
        } catch (IOException e){
            spectraLogger.info(e.getMessage());
            e.printStackTrace();
        }

        timeSinceLastSearch = System.currentTimeMillis();
        return "Complete";
    }


    // Initializes message trackers for each thread
    private void getMessageTrackers() {
        MessageTracker msgT[] = new MessageTracker[numberOfWorkSpaceThreads];
        for (int i = 0; i < numberOfWorkSpaceThreads; i++) {
            msgT[i] = new MessageTracker();
        }
        this.msgTrackersSqt = msgT;
        this.msgTrackerMs2 = new MessageTracker();
    }


    /**
     *
     * @return a buffered writer array of size equal to the number of threads
     */
    private void getBufferedWriters() {

        //Make folder for producerID to store sqt files
        BufferedWriter bw[] = new BufferedWriter[numberOfWorkSpaceThreads];

        try{
            for(int i = 0; i < numberOfWorkSpaceThreads; i++){
                FileWriter fw = new FileWriter(outputDir + File.separator  + "thread" + i +".sqt");
                bw[i] = new BufferedWriter(fw);
            }
            this.bufWritersSqt = bw;

            FileWriter fw = new FileWriter(outputDir+ File.separator + uniqueId +".ms2");
            this.bufWriterMs2 = new BufferedWriter(fw);
            this.bufWriterMs2.write(
                    "H       Extractor       TimsTOF_RTMS\n" +
                            "H       AcquisitionMethod       Data-Dependent\n" +
                            "H       InstrumentType  TIMSTOF\n" +
                            "H       DataType        Centroid\n" +
                            "H       ScanType        MS2\n" +
                            "H       MonoIsotopic PrecMz     TrueH       Extractor       TimsTOF_RTMS\n" +
                            "H       AcquisitionMethod       Data-Dependent\n" +
                            "H       InstrumentType  TIMSTOF\n" +
                            "H       DataType        Centroid\n" +
                            "H       ScanType        MS2\n" +
                            "H       MonoIsotopic PrecMz     True\n");

        } catch (IOException e) {
            spectraLogger.error("An error occurred creating BufferWriters.");
            spectraLogger.error(e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * close all buffered writers to avoid losing data
     */
    private void closeBufferWriters(){

        if (bufWritersSqt == null){
            return;
        }

        try {
            for (int i = 0; i < numberOfWorkSpaceThreads; i++) {
                if (bufWritersSqt[i] != null) {
                    bufWritersSqt[i].close();
                }
            }
            bufWriterMs2.close();
        } catch (IOException e){
            spectraLogger.error("An error occurred closing fileWriters.");
            spectraLogger.error(e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * synchronized method for assign unique id to thread name
     * @param threadName
     */
    public synchronized void setUpThreadForGpuSearch(String threadName){

        //Find gpuWorkSpace which isn't assigned to thread
        int numberOfThreadsAssigned = uniqueThreadIndexHashTable.size();
        uniqueThreadIndexHashTable.put(threadName, numberOfThreadsAssigned);

    }

    /**
     * free all gpu memory, once not workspace has searched in X amount of time
     * close all buffered writers
     */
    public void free() {
        spectraLogger.info("Stopping Scheduler bc End sequence");
        try {
            while (System.currentTimeMillis() - timeSinceLastSearch < waitTimeBeforeThreadIsConsideredFinished) {
                spectraLogger.info("(SQT) " + Thread.currentThread());
                Thread.sleep(1000);
            }

            spectraLogger.info("Dispose kafka Receiver and scheduler, close bufferedWriters");
            kafkaReceivorDisposable.dispose();
            databaseSearchResultSender.close();
            databaseSeachScheduler.dispose();
            ms2FileScheduler2.dispose();

            closeBufferWriters();
            if(use_gpu) prolucidController.free();

        } catch (InterruptedException e){
            spectraLogger.error("Exception freeing ms2Consumer");
            spectraLogger.error(e.getMessage());
            e.printStackTrace();
        }

        // Handle message trackers / missed messages
        MessageTracker combined = new MessageTracker();
        for(MessageTracker ms : msgTrackersSqt){
            combined.combine(ms);
        }

        combined.printMissedMessages("Ms2 DB");
        spectraLogger.info("[Ms2 DB] totalMessagesReceived: " + combined.numMsg);
        spectraLogger.info("[Ms2 DB] totalBytesReceived: " + combined.totalSizeOfMsg);

        msgTrackerMs2.printMissedMessages("MS2 File");
        spectraLogger.info("[MS2 File] totalMessagesReceived: " + msgTrackerMs2.numMsg);
        spectraLogger.info("[MS2 File] totalBytesReceived: " + msgTrackerMs2.totalSizeOfMsg);
        spectraLogger.info("[MS2 File] number_ofSpectra_received: " + number_ofSpectra_received_ms2);

    }

    private ReceiverOptions<String, byte[]> receiverOptions() {
        Map<String, Object> props = new HashMap<>();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, paserConstants.getKafkaIP());
        props.put(ConsumerConfig.GROUP_ID_CONFIG, receiverGroupID);
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG,
                StringDeserializer.class.getName());
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG,
                ByteArrayDeserializer.class.getName());
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "latest");
        return ReceiverOptions.create(props);
    }

    private ReceiverOptions<String, byte[]> receiverOptions(Collection<String> topic) {
        return receiverOptions()
                //.commitInterval(Duration.ofSeconds(10))
                .addAssignListener(partitions -> spectraLogger.info("onPartitionsAssigned {}" + partitions))
                .addRevokeListener(partitions -> spectraLogger.info("onPartitionsRevoked {}" + partitions))
                .subscription(topic);
    }

    private SenderOptions<String, String> senderOptions() {
        Map<String, Object> senderProps = new HashMap<>();
        senderProps.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, paserConstants.getKafkaIP());
        senderProps.put(ProducerConfig.CLIENT_ID_CONFIG, senderGroupID);
        senderProps.put(ProducerConfig.ACKS_CONFIG, "all");
        senderProps.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG,
                StringSerializer.class.getName());
        senderProps.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG,
                StringSerializer.class.getName());
        return SenderOptions.create(senderProps);
    }

    private KafkaSender<String, String> sender(SenderOptions<String, String> senderOptions) {
        return KafkaSender.create(senderOptions);
    }
}