package com.ip2.paser;

import com.ip2.paser.util.PaserConstants;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.common.serialization.ByteArrayDeserializer;
import org.apache.kafka.common.serialization.StringDeserializer;
import reactor.core.Disposable;
import reactor.core.scheduler.Scheduler;
import reactor.core.scheduler.Schedulers;
import reactor.kafka.receiver.KafkaReceiver;
import reactor.kafka.receiver.ReceiverOffset;
import reactor.kafka.receiver.ReceiverOptions;
import reactor.kafka.receiver.ReceiverRecord;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.time.Duration;
import java.util.*;

public class PrecursorConsumer {

    private final ReceiverOptions<String, byte[]> receiverOptions;
    ReceiverOptions<String, byte[]> options;

    BufferedWriter bufWriter;
    String uniqueId;
    Scheduler scheduler;

    long waitTimeBeforeThreadIsConsideredFinished = 5000;
    long timeSinceLastSearch;
    Disposable kafkaReceivorDisposable;

    int scanNumberItr = 0;
    MessageTracker msgTrack;

    private String outputDir;

    public int totalMessagesReceived = 0;
    public int totalBytesReceived = 0;

    private PaserConstants paserConstants;


    /**
     * Reactive Kafka Consumer to handle writing the raw data to an ms2 file
     * @param id the unique id associated with this connection to the instrument
     */
    public PrecursorConsumer(String id) {

        paserConstants = PaserConstants.getInstance();

        uniqueId = id;
        msgTrack = new MessageTracker();
        timeSinceLastSearch = System.currentTimeMillis();
        outputDir = paserConstants.getDataOutPath() + "/" + uniqueId;


        try{
            FileWriter fw = new FileWriter(outputDir+"/Precursors_" + uniqueId +".txt");
            bufWriter = new BufferedWriter(fw);
        } catch (IOException e) {
            System.out.println("An error occurred creating BufferWriters.");
            e.printStackTrace();
        }

        Map<String, Object> props = new HashMap<>();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, paserConstants.getKafkaIP());
        props.put(ConsumerConfig.GROUP_ID_CONFIG, "spectra-consumer-precursor");
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG,
                StringDeserializer.class.getName());
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG,
                ByteArrayDeserializer.class.getName());
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "latest");
        this.receiverOptions = ReceiverOptions.create(props);
    }


    /**
     * Reactive consumer to handle Spectra messages.
     */
    public void consumeMessages() {
        System.out.println("Starting Method: com.ip2.paser.Precursor Consumer");
        scheduler = Schedulers.newSingle("com.ip2.paser.PrecursorConsumer", false);

        options = receiverOptions.subscription(Collections.singleton("com.ip2.paser.Precursor"))
                .addAssignListener(partitions -> System.out.println("onPartitionsAssigned {}" + partitions))
                .addRevokeListener(partitions -> System.out.println("onPartitionsRevoked {}" + partitions));

        kafkaReceivorDisposable = KafkaReceiver.create(options.commitInterval(Duration.ZERO))
                .receive()
                .groupBy(m -> 1)
                .flatMap(partitionFlux -> partitionFlux.publishOn(scheduler)
                        .map(r -> processRecord(r))
                        .concatMap(offset -> {
                            offset.acknowledge();
                            System.out.println("Commit: com.ip2.paser.Precursor Consumer");
                            return offset.commit();
                        })).subscribe();

        System.out.println("Ending Method: com.ip2.paser.Precursor Consumer");
    }

    /**
     * Scheduler calls this method to run on each threads. Searches spectra on gpuWorksSpace
     * @param message Spectra message
     * @return return isn't used
     */
    public ReceiverOffset processRecord(ReceiverRecord<String, byte[]> message) {

        System.out.println("processRecord com.ip2.paser.PrecursorConsumer");

        String msgId;
        int ms1_frame_number;

        String keySplit[] = message.key().split(" ");
        msgId = keySplit[0];
        ms1_frame_number = Integer.parseInt(keySplit[1]);

        //System.out.println("Process record Sqt");
        if(msgId.compareTo(uniqueId) != 0 ){
            System.out.println("[com.ip2.paser.PrecursorConsumer] Skipping spectra for another id: " + msgId + ", MsgNum: " +
                    ms1_frame_number + ", current ID: " + uniqueId);
            return message.receiverOffset();
        }

        msgTrack.add(message, ms1_frame_number);

        PrecursorScanParser precScan = new PrecursorScanParser(message.value());
        precScan.print();

        try{
            for ( Precursor p : precScan.pList){
                bufWriter.write(p.toString());
            }
        } catch (IOException e){
            System.out.println("An error occurred writing to BufferWriters.");
            e.printStackTrace();
        }

        timeSinceLastSearch = System.currentTimeMillis();
        return message.receiverOffset();
    }


    /**
     * free all gpu memory, once not workspace has searched in X amount of time
     * close all buffered writers
     */
    public void free() {
        System.out.println("Stopping Scheduler bc End sequence");
        try {
            while (System.currentTimeMillis() - timeSinceLastSearch < waitTimeBeforeThreadIsConsideredFinished) {
                System.out.println("(MS2) " + Thread.currentThread());
                Thread.sleep(1000);
            }
            System.out.println("Dispose kafka Receiver and scheduler, close bufferedWriters");
            kafkaReceivorDisposable.dispose();
            scheduler.dispose();

            try{
                bufWriter.close();
            } catch (IOException e){
                System.out.println(e.getStackTrace());
                System.out.println(e);
            }
        } catch (InterruptedException e){
            System.out.println(e);
        }
        msgTrack.printMissedMessages("[com.ip2.paser.PrecursorConsumer]");
        totalMessagesReceived = msgTrack.numMsg;
        totalBytesReceived = msgTrack.totalSizeOfMsg;
        System.out.println("[com.ip2.paser.PrecursorConsumer] totalMessagesReceived: " + totalMessagesReceived);
        System.out.println("[com.ip2.paser.PrecursorConsumer] totalBytesReceived: " + totalBytesReceived);

    }
}
