package com.ip2.paser;
import edu.scripps.pms.util.spectrum.Peak;
import edu.scripps.pms.util.spectrum.PeakList;
import edu.scripps.pms.util.spectrum.Zline;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.DoubleBuffer;
import java.util.Iterator;


public class SpectraScanParser {

    double PROTON_MASS = 1.00727826;

    public float scanNumber;
    public int charge;
    public float analysisTime;
    public float precursorMass;
    public float intensity;
    public int ms2FrameNumber;
    public int ms1FrameNumber;
    public int Ms2SlineItr;
    public PeakList pList;
    public float massInDaltons;

    public SpectraScanParser(byte[] rawData) {
        pList = new PeakList();

        parseSpectra(rawData);
    }

    /**
     * Parses Raw spectra bytes into scanNumber, charge, analysisTime, precursorMass, and peakList
     * @param rawData
     */
    private void parseSpectra(byte[] rawData) {

        DoubleBuffer fBuf = ByteBuffer.wrap(rawData)
                .order(ByteOrder.LITTLE_ENDIAN)
                .asDoubleBuffer();
        double[] spectraArray = new double[fBuf.remaining()];
        fBuf.get(spectraArray);

        if(spectraArray.length < 4){
            System.out.println("Error: Length of rawData is: " + rawData.length);
            return;
        }

        scanNumber = (float)spectraArray[0];
        charge = (int)spectraArray[1];
        analysisTime = (float)spectraArray[2];
        precursorMass = (float)spectraArray[3];
        intensity = (float)spectraArray[4];
        ms2FrameNumber = (int)spectraArray[5];
        Ms2SlineItr = (int)spectraArray[6];
        ms1FrameNumber = (int)spectraArray[7];



        // Unknown charge state, try with 2 and 3
        if(charge == 0){
            pList.addZline(new Zline(2, (float)((precursorMass - PROTON_MASS) * 2 + PROTON_MASS)));
            pList.addZline(new Zline(3, (float)((precursorMass - PROTON_MASS) * 3 + PROTON_MASS)));
            //pList.addZline(new Zline(2, precursorMass));
            //pList.addZline(new Zline(3, precursorMass));
            massInDaltons = precursorMass;


        }else {
            massInDaltons = (float)((precursorMass - PROTON_MASS) * charge + PROTON_MASS);
            massInDaltons = (float) (roundToDecimalPlace(massInDaltons,4));

            pList.addZline(new Zline(charge, massInDaltons));
            //pList.addZline(new Zline(charge, precursorMass));

        }

        for (int i = 8; i < spectraArray.length - 1; i += 2) {

            //double mzValue = spectraArray[i+1];
            double intensityValue = spectraArray[i];

            double mzValue = roundToDecimalPlace(spectraArray[i+1],4);

            if(mzValue != 0.0 && intensityValue != 0.0) {
                pList.addPeak(new Peak(mzValue, intensityValue));
            }
        }
        pList.setPrecursorMass(roundToDecimalPlace(precursorMass, 4));
        //pList.addIline("I" + "\t" + pList.PRECURSORSCAN + "\t" + Ms2SlineItr );
        //pList.addIline("I" + "\t" + pList.RETENTIONTIME + "\t" + analysisTime );
        pList.setLoscan(Ms2SlineItr);
        pList.setHiscan(Ms2SlineItr);
    }

    private double roundToDecimalPlace(double value, int numberOfDecimalValuesToShow){
        double divisor = 1.0*numberOfDecimalValuesToShow;
        return Math.round(value*divisor)/divisor;
    }

    public PeakList getPeakList() {
        return pList;
    }

    public void print() {
        System.out.println("ScanNumber: " + scanNumber);
        System.out.println("    charge: " + charge);
        System.out.println("    analysisTime: " + analysisTime);
        System.out.println("    precurserMass: " + precursorMass);
        System.out.println("    Number of Peaks: " + pList.numPeaks());
        System.out.println();

    }

    public String debugString(){
        StringBuilder sb = new StringBuilder();

        String sLine = "S\t" + Ms2SlineItr + "\t" + Ms2SlineItr +"\t"+ String.format("%.4f",precursorMass) +"\n";
        String iLine1 = "I\tTIMSTOF_Parent_MSMS_Frame\t" + ms1FrameNumber +"\n";
        String iLine2 = "I\tTIMSTOF_Precursor_ID\t" +"?? \n";
        String iLine3 = "I\tRetTime\t" + analysisTime+"\n";
        double prec_mass_daltons = ((precursorMass - PROTON_MASS) * charge + PROTON_MASS);
        String zLine = "Z\t" + charge + "\t" + prec_mass_daltons+"\n";

        sb.append(sLine);
        sb.append(iLine1);
        sb.append(iLine2);
        sb.append(iLine3);
        sb.append(zLine);

        return sb.toString();
    }

    public String Ms2ToString(){
        StringBuilder sb = new StringBuilder();

        String sLine = "S\t" + Ms2SlineItr + "\t" + Ms2SlineItr +"\t"+ String.format("%.4f", precursorMass) +"\n";
        String iLine1 = "I\tTIMSTOF_Parent_MSMS_Frame\t" + ms1FrameNumber +"\n";
        String iLine2= "I\tRetTime\t" + String.format("%.4f",analysisTime) +"\n";
        String zLine = "Z\t" + charge + "\t" + String.format("%.4f",massInDaltons) +"\n";

        sb.append(sLine);
        sb.append(iLine1);
        sb.append(iLine2);
        sb.append(zLine);

        Iterator<Peak> pItr = pList.getPeaks();
        while(pItr.hasNext()){
            Peak p = pItr.next();
            sb.append(String.format("%.4f", roundToDecimalPlace(p.getM2z(),4)) + " " + p.getIntensity() + "\n");
            //sb.append(p.getIntensity() + " " + p.getM2z() + "\n");
        }

        return sb.toString();

    }
}